"""Logic for compiling requirements"""
import logging

import pkg_resources

import qer.metadata
import qer.pypi
from qer import utils
from qer.utils import merge_requirements, normalize_project_name


class MetadataSources(object):
    def __init__(self, metadata, source):
        self.metadata = metadata
        self.sources = {source}


class DistributionCollection(object):
    CONSTRAINTS_ENTRY = '#constraints#'

    def __init__(self, constraints=None):
        self.dists = {}
        constraints_dist = qer.metadata.DistInfo()
        constraints_dist.name = '#constraints#'
        constraints_dist.reqs = constraints or []
        constraints_dist.version = None
        self.constraints_dist = constraints_dist
        self.dists[DistributionCollection.CONSTRAINTS_ENTRY] = MetadataSources(
            constraints_dist, DistributionCollection.CONSTRAINTS_ENTRY)
        self.orig_roots = None

    def add_dist(self, metadata, source):
        if metadata.name in self.dists:
            self.dists[normalize_project_name(metadata.name)].sources.add(source)
        else:
            self.dists[normalize_project_name(metadata.name)] = MetadataSources(metadata, source)

    def remove_dist(self, name):
        del self.dists[normalize_project_name(name)]

    def remove_source(self, source):
        dists_to_remove = []
        for dist in self.dists.itervalues():
            if source in dist.sources:
                dist.sources.remove(source)
                if not dist.sources:
                    dists_to_remove.append(normalize_project_name(dist.metadata.name))

        for dist in dists_to_remove:
            del self.dists[dist]

    def __contains__(self, item):
        return normalize_project_name(item) in self.dists

    def add_global_constraint(self, constraint):
        self.constraints_dist.reqs.append(constraint)

    def build_constraints(self, project_name):
        normalized_name = normalize_project_name(project_name)
        req = None if normalized_name == DistributionCollection.CONSTRAINTS_ENTRY \
            else utils.parse_requirement(normalized_name)
        for dist_name, dist in self.dists.iteritems():
            for subreq in dist.metadata.reqs:
                if normalize_project_name(subreq.name) == normalized_name:
                    req = merge_requirements(req, subreq)
                    break
        return req

    def get_reverse_deps(self, project_name):
        reverse_deps = {}
        normalized_name = normalize_project_name(project_name)
        for dist_name, dist in self.dists.iteritems():
            for subreq in dist.metadata.reqs:
                if normalize_project_name(subreq.name) == normalized_name:
                    reverse_deps[dist_name] = subreq
                    break
        return reverse_deps


def compile_roots(root, source, extras=(), dists=None, round=1, index_url=None,
                  toplevel=None, session=None, wheeldir=None):
    logger = logging.getLogger('qer.compile')

    if not qer.metadata.filter_req(root, extras):
        return

    specifier = dists.build_constraints(root.name).specifier
    if root.name in dists:
        normalized_name = normalize_project_name(root.name)
        logger.info('Reusing dist %s %s', root.name, dists.dists[normalized_name].metadata.version)
        if not specifier.contains(dists.dists[normalized_name].metadata.version):
            raise EnvironmentError('Already existing dist did not match constraint: {}'.format(specifier))
        dists.dists[normalized_name].sources.add(source)
        metadata = dists.dists[normalized_name].metadata
    else:
        try:
            dist = qer.pypi.download_candidate(root.name, specifier=specifier,
                                               index_url=index_url, session=session, wheeldir=wheeldir)
        except qer.pypi.NoCandidateException as ex:
            logger.info('No candidate for %s. Contributions: %s',
                        ex.project_name, dists.get_reverse_deps(ex.project_name))
            raise

        metadata = qer.metadata.extract_metadata(dist, extras=extras)
        dists.add_dist(metadata, source)

        # See how the new constraints do with the already collected reqs
        for dist in dists.dists.values():
            if dist.metadata.name != DistributionCollection.CONSTRAINTS_ENTRY:
                constraints = dists.build_constraints(dist.metadata.name)
                if not constraints.specifier.contains(dist.metadata.version):
                    logger.info('Already selected dist violated (%s %s)',
                                 dist.metadata.name, dist.metadata.version)

                    # Remove all downstream reqs
                    dists.remove_source(dist.metadata.name)
                    dists.remove_dist(dist.metadata.name)

                    dists.add_global_constraint(utils.parse_requirement(
                        '{}!={}'.format(dist.metadata.name, dist.metadata.version)))
                    return compile_roots(toplevel, 'rerun',
                                         extras=extras, dists=dists, round=round+1,
                                         toplevel=toplevel, index_url=index_url, session=session,
                                         wheeldir=wheeldir)

    for req in metadata.reqs:
        compile_roots(req, normalize_project_name(root.name), dists=dists, round=round,
                      toplevel=toplevel, index_url=index_url, session=session, wheeldir=wheeldir)



import itertools
import os
from collections import defaultdict
import logging

import functools32
import pkg_resources


def _req_iter_from_file(reqfile_name):
    with open(reqfile_name, 'r') as reqfile:
        for req_line in reqfile:
            req_line = req_line.strip()
            if not req_line:
                continue
            if req_line.startswith('-r'):
                for req in _req_iter_from_file(os.path.join(
                        os.path.dirname(reqfile_name),
                        req_line.split(' ')[1].strip())):
                    yield req
            elif req_line.startswith('--index-url') or req_line.startswith('--extra-index-url'):
                pass
            elif req_line.startswith('#'):
                pass
            else:
                try:
                    yield parse_requirement(req_line)
                except ValueError:
                    logging.getLogger('req.utils').exception('Failed to parse %s', req_line)
                    raise


def reqs_from_files(requirements_files):
    """
    Args:
        requirements_files (list[str]): Requirements files
    """
    raw_reqs = iter([])
    for reqfile_name in requirements_files:
        raw_reqs = itertools.chain(raw_reqs, _req_iter_from_file(reqfile_name))

    reqs = defaultdict(lambda: None)
    for req in raw_reqs:
        reqs[req.name] = merge_requirements(reqs[req.name], req)

    return list(reqs.values())


@functools32.lru_cache(maxsize=1500)
def parse_requirement(req_text):
    return pkg_resources.Requirement.parse(req_text)


def parse_requirements(reqs):
    for req in reqs:
        yield parse_requirement(req)


@functools32.lru_cache(maxsize=1000)
def merge_requirements(req1, req2):
    if req1 is not None and req2 is None:
        return req1
    if req2 is not None and req1 is None:
        return req2

    assert normalize_project_name(req1.name) == normalize_project_name(req2.name)
    all_specs = set(req1.specs or []) | set(req2.specs or [])
    if req1.marker and req2.marker and str(req1.marker) != str(req2.marker):
        if str(req1.marker) in str(req2.marker):
            new_marker = ';' + str(req2.marker)
        elif str(req2.marker) in str(req1.marker):
            new_marker = ';' + str(req1.marker)
        else:
            new_marker = ';' + str(req1.marker) + ' and ' + str(req2.marker)
    elif req1.marker:
        new_marker = ';' + str(req1.marker)
    elif req2.marker:
        new_marker = ';' + str(req2.marker)
    else:
        new_marker = ''

    req_str = normalize_project_name(req1.name) + ','.join(''.join(parts) for parts in all_specs) + new_marker
    return parse_requirement(req_str)


@functools32.lru_cache(maxsize=500)
def normalize_project_name(project_name):
    return project_name.lower().replace('-', '_').replace('.', '_')



import collections
import logging
import os
import shutil
import tempfile
import urlparse
from HTMLParser import HTMLParser

import pkg_resources
import requests
from hashlib import sha256
import functools32

Candidate = collections.namedtuple('Candidate', 'name filename version py_version link')


logger = logging.getLogger('qer.pypi')


@functools32.lru_cache()
def _scan_page_links(index_url, project_name, session):
    """

    Args:
        index_url:
        project_name:
        session (requests.Session): Session

    Returns:
        (list[Candidate])
    """
    url = '{index_url}/{project_name}'.format(index_url=index_url, project_name=project_name)
    logging.getLogger('qer.net.pypi').info('Fetching versions for %s', project_name)
    if session is None:
        session = requests
    response = session.get(url + '/')

    class LinksHTMLParser(HTMLParser):
        def __init__(self, url):
            HTMLParser.__init__(self)
            self.url = url
            self.dists = []
            self.active_link = None

        def handle_starttag(self, tag, attrs):
            if tag == 'a':
                for attr in attrs:
                    if attr[0] == 'href':
                        self.active_link = urlparse.urljoin(self.url, attr[1])
                        break

        def handle_data(self, filename):
            extensions = ('.whl', '.tar.gz', '.tgz', '.zip')
            if '.whl' in filename:
                data_parts = filename.split('-')
                name = data_parts[0]
                version = pkg_resources.parse_version(data_parts[1])
                abi = data_parts[3]
                platform = data_parts[4].split('.')[0]
                if platform == 'any' or platform == 'win_amd64':
                    self.dists.insert(0, Candidate(name,
                                                   filename,
                                                   version,
                                                   tuple(data_parts[2].split('.')),
                                                   self.active_link))
            elif '.tar.gz' in filename or '.tgz' in filename or '.zip' in filename:
                data_parts = filename.split('-')
                name = data_parts[0]
                version_text = data_parts[-1]
                for ext in extensions:
                    version_text = version_text.replace(ext, '')
                version = pkg_resources.parse_version(version_text)
                self.dists.insert(0, Candidate(name, filename, version, (), self.active_link))

    parser = LinksHTMLParser(response.url)
    parser.feed(response.content)

    return sorted(parser.dists, key=lambda x: x.version, reverse=True)


def _do_download(index_url, filename, link, session, wheeldir):
    split_link = link.split('#sha256=')
    sha = split_link[1]

    output_file = os.path.join(wheeldir, filename)

    if os.path.exists(output_file):
        hasher = sha256()
        with open(output_file, 'rb') as handle:
            while True:
                block = handle.read(1024)
                if not block:
                    break
                hasher.update(block)
        if hasher.hexdigest() == sha:
            logger.info('Reusing %s', output_file)
            return output_file

        print "File hash doesn't match"

    full_link = split_link[0]
    logging.getLogger('qer.net.pypi').info('Downloading %s -> %s', full_link, output_file)
    if session is None:
        session = requests
    response = session.get(full_link, stream=True)

    with open(output_file, 'wb') as handle:
        for block in response.iter_content(4 * 1024):
            handle.write(block)
    return output_file


class NoCandidateException(Exception):
    def __init__(self, *args):
        super(NoCandidateException, self).__init__(*args)
        self.project_name = None
        self.specifier = None

    def __str__(self):
        return 'NoCandidateException - no candidate for "{}" satisfies {}'.format(
            self.project_name,
            self.specifier
        )


def start_session():
    return requests.Session()


def download_candidate(project_name, py_ver='py2', specifier=None, allow_prerelease=False,
                       index_url=None, skip_source=True, session=None, wheeldir=None):
    logger = logging.getLogger('qer.download')
    logger.info('Downloading %s, with constraints %s', project_name, specifier)
    if index_url is None:
        index_url = 'https://pypi.org/simple'
    candidates = _scan_page_links(index_url, project_name, session)

    delete_wheeldir = False
    if wheeldir is None:
        wheeldir = tempfile.mkdtemp()
        delete_wheeldir = True

    for candidate in candidates:
        if not ('py2' in candidate.py_version or 'cp27' in candidate.py_version) and skip_source:
            continue

        has_equality = any(spec.operator == '==' for spec in specifier)

        if not has_equality and not allow_prerelease and candidate.version.is_prerelease:
            continue

        if specifier is not None and not specifier.contains(candidate.version):
            continue

        return _do_download(index_url, candidate.filename, candidate.link, session, wheeldir)

    if not skip_source:
        ex = NoCandidateException()
        ex.project_name = project_name
        ex.specifier = specifier
        raise ex

    return download_candidate(project_name, py_ver=py_ver, specifier=specifier,
                              allow_prerelease=allow_prerelease, index_url=index_url,
                              skip_source=False, session=session, wheeldir=wheeldir)



import argparse
import logging
from collections import defaultdict

from hashlib import sha256

from qer.utils import merge_requirements
import qer.utils


def run_hash(requirements_files):
    """
    Args:
        requirements_files (list[str]): Requirements files
    """
    reqs = defaultdict(lambda: None)
    for req in qer.utils.reqs_from_files(requirements_files):
        reqs[req.name] = merge_requirements(reqs[req.name], req)

    hasher = sha256()
    for req in reqs.values():
        hasher.update(str(req))
    return hasher.hexdigest()


def hash_main():
    logging.basicConfig(level=logging.DEBUG)

    parser = argparse.ArgumentParser()
    parser.add_argument('requirement_files', nargs='+', help='Input requirements files')

    args = parser.parse_args()
    print run_hash(args.requirement_files)


if __name__ == '__main__':
    hash_main()



import os
import tarfile
import zipfile
import functools

import pkg_resources

from qer import pypi, utils


class DistInfo(object):
    def __init__(self):
        self.reqs = []
        self.name = None
        self.version = None

    def __repr__(self):
        return self.name + ' ' + self.version + '\n' + '\n'.join([str(req) for req in self.reqs])


def filter_req(req, extras):
    keep_req = True
    if req.marker:
        if extras:
            keep_req = any(req.marker.evaluate({'extra': extra}) for extra in extras)
        else:
            keep_req = req.marker.evaluate({'extra': None})
    return keep_req


def extract_metadata(dist, extras=()):
    """"""
    if dist.lower().endswith('.whl'):
        return _fetch_from_wheel(dist, extras=extras)
    if dist.lower().endswith('.zip'):
        return _fetch_from_zip(dist, extras=extras)
    elif dist.lower().endswith('.tar.gz'):
        return _fetch_from_source(dist, extras=extras)


def _fetch_from_zip(zip_file, extras):
    zfile = zipfile.ZipFile(zip_file, 'r')
    try:
        metadata_file = None
        pkg_info_file = None
        egg_info = None

        for name in zfile.namelist():
            if name.lower().endswith('pkg-info'):
                pkg_info_file = name
            elif name.lower().endswith('.egg-info'):
                egg_info = name
            elif name.lower().endswith('metadata'):
                metadata_file = name

        if egg_info:
            filename = os.path.basename(zip_file)
            name = '-'.join(filename.split('-')[0:-1])
            version = pkg_resources.parse_version(filename.split('-')[-1].replace('.tar.gz', ''))
            return _parse_requires_file(zip_file.extractfile(egg_info + '/requires.txt').read(),
                                        name,
                                        version,
                                        extras)

        if pkg_info_file:
            return _parse_flat_metadata(zfile.read(pkg_info_file), extras)

        if metadata_file:
            return _parse_flat_metadata(zfile.read(metadata_file), extras)
    finally:
        zfile.close()


def _fetch_from_source(tar_gz, extras):
    tar = tarfile.open(tar_gz, "r:gz")
    try:
        metadata_file = None
        pkg_info_file = None
        egg_info = None

        for info in tar.getmembers():
            name = info.name.lower()
            if name.endswith('pkg-info'):
                pkg_info_file = info.name
            elif name.endswith('.egg-info'):
                egg_info = info.name
            elif name.endswith('metadata'):
                metadata_file = info.name

        if egg_info:
            filename = os.path.basename(tar_gz)
            name = '-'.join(filename.split('-')[0:-1])
            version = pkg_resources.parse_version(filename.split('-')[-1].replace('.tar.gz', ''))
            requires_contents = ''
            try:
                requires_contents = tar.extractfile(egg_info + '/requires.txt').read()
            except KeyError:
                pass
            return _parse_requires_file(requires_contents,
                                        name,
                                        version,
                                        extras)

        if pkg_info_file:
            return _parse_flat_metadata(tar.extractfile(pkg_info_file).read(), extras)

        if metadata_file:
            return _parse_flat_metadata(tar.extractfile(metadata_file).read(), extras)
    finally:
        tar.close()


def _fetch_from_wheel(wheel, extras):
    zfile = zipfile.ZipFile(wheel, 'r')
    try:
        metadata_file = None
        infos = zfile.namelist()
        for info in infos:
            if info.lower().endswith('metadata'):
                metadata_file = info

        if metadata_file:
            return _parse_flat_metadata(zfile.read(metadata_file), extras)
    finally:
        zfile.close()


def _parse_flat_metadata(contents, extras):
    result = DistInfo()
    raw_reqs = []
    for line in contents.split('\n'):
        if line.lower().startswith('name:'):
            result.name = line.split(':')[1].strip()
        if line.lower().startswith('version:'):
            result.version = pkg_resources.parse_version(line.split(':')[1].strip())
        if line.lower().startswith('requires-dist:'):
            raw_reqs.append(line.split(':')[1].strip())

    result.reqs = [req for req in utils.parse_requirements(raw_reqs)
                   if filter_req(req, extras)]
    return result


def _parse_requires_file(contents, name, version, extras):
    result = DistInfo()
    reqs = []
    sections  = list(pkg_resources.split_sections(contents))
    for section in sections:
        if section[0] is None:
            reqs.extend(utils.parse_requirements(section[1]))
        elif section[0].startswith(':python_version'):
            for req in section[1]:
                reqs.append(utils.parse_requirement(req + ' ' + section[0].replace(':', ';')))
    result.reqs = reqs
    result.name = name
    result.version = version
    return result



from __future__ import print_function

import argparse
import logging
import os
import shutil
import string
import sys
import tempfile
from contextlib import closing

import pkg_resources

import qer.compile
import qer.metadata
import qer.pypi
from qer import utils

ROOT_REQ = 'root__a'

BLACKLIST = [
    qer.compile.DistributionCollection.CONSTRAINTS_ENTRY,
    'setuptools'
]


def _get_reason_constraint(dists, constraint_dists, project_name, root_mapping):
    project_name = utils.normalize_project_name(project_name)
    components = dists.get_reverse_deps(project_name)
    if not components:
        constraints = ''
    else:
        constraints = []
        for component in components:
            if component == ROOT_REQ:
                continue
            for req in dists.dists[component].metadata.reqs:
                if utils.normalize_project_name(req.name) == project_name:
                    if component == qer.compile.DistributionCollection.CONSTRAINTS_ENTRY:
                        constraints_reason = _get_reason_constraint(constraint_dists, None,
                                                                    project_name, root_mapping)
                        if constraints_reason:
                            constraints += ['(via constraints: ' + constraints_reason + ')']
                    else:
                        specifics = ' (' + str(req.specifier) + ')' if req.specifier else ''
                        if component.startswith(ROOT_REQ):
                            component = root_mapping[component]
                        constraints += [component + specifics]
                        break
        constraints = ', '.join(constraints)
    return constraints


def _generate_lines(dists, constraint_dists, root_mapping):
    for dist in dists.dists.itervalues():
        if dist.metadata.name in BLACKLIST:
            continue
        if dist.metadata.name.startswith(ROOT_REQ):
            continue

        constraints = _get_reason_constraint(dists, constraint_dists, dist.metadata.name, root_mapping)

        constraint = '{}=={}'.format(dist.metadata.name, dist.metadata.version).ljust(43)
        yield '{}# {}'.format(constraint, constraints)


def _generate_constraints(dists):
    for dist in dists.dists.itervalues():
        if dist.metadata.name in BLACKLIST:
            continue
        if dist.metadata.name.startswith(ROOT_REQ):
            continue

        req = dists.build_constraints(dist.metadata.name)
        if req.specifier:
            yield req


def _generate_no_candidate_display(ex, dists, constraint_dists, root_mapping):
    project_name = utils.normalize_project_name(ex.project_name)
    components = dists.get_reverse_deps(project_name)
    if not components:
        print('No package available for %s, latest is %s', file=sys.stderr)
    else:
        print('No version of {} could satisfy the following requirements:'.format(ex.project_name), file=sys.stderr)
        for component in components:
            for req in dists.dists[component].metadata.reqs:
                if utils.normalize_project_name(req.name) == project_name:
                    if component == qer.compile.DistributionCollection.CONSTRAINTS_ENTRY:
                        constraints_reason = _get_reason_constraint(constraint_dists, None,
                                                                    project_name, root_mapping)
                        if constraints_reason:
                            print('   ' + constraints_reason, file=sys.stderr)
                    else:
                        specifics = str(req.specifier) if req.specifier else ''
                        if component.startswith(ROOT_REQ):
                            source = root_mapping.get(component, 'input file')
                        else:
                            constraints_reason = _get_reason_constraint(dists, constraint_dists,
                                                                        component, root_mapping)
                            if constraints_reason:
                                constraints_reason = ' (via ' + constraints_reason + ')'
                            source = '{} {}{}'.format(component, dists.dists[component].metadata.version, constraints_reason)

                        print('   {} requires {}{}'.format(source,
                                                           ex.project_name, specifics),
                              file=sys.stderr)
                        break
    sys.exit(1)


def _build_root_metadata(roots, name):
    metadata = qer.metadata.DistInfo()
    metadata.name = name
    metadata.version = '0'
    metadata.reqs = list(roots)
    return metadata


def run_compile(input_reqfiles, constraint_files, index_url, wheeldir, no_combine):
    root_req = utils.parse_requirement(ROOT_REQ)

    if wheeldir:
        if not os.path.exists(wheeldir):
            os.mkdir(wheeldir)
        delete_wheeldir = False
    else:
        wheeldir = tempfile.mkdtemp()
        delete_wheeldir = True

    constraints = None
    constraint_results = qer.compile.DistributionCollection()

    if constraint_files:
        constraint_roots = utils.reqs_from_files(constraint_files)

        constraint_results = qer.compile.DistributionCollection()
        constraint_results.add_dist(_build_root_metadata(constraint_roots, ROOT_REQ), ROOT_REQ)

        with closing(qer.pypi.start_session()) as session:
            qer.compile.compile_roots(root_req, ROOT_REQ, dists=constraint_results,
                                      toplevel=root_req, index_url=index_url, session=session,
                                      wheeldir=wheeldir)

        constraints = list(_generate_constraints(constraint_results))

    results = qer.compile.DistributionCollection(constraints)
    root_mapping = {}
    try:
        if no_combine:
            fake_reqs = []
            idx = 1
            for input_reqfile in input_reqfiles:
                roots = utils.reqs_from_files([input_reqfile])
                name = '{}{}'.format(ROOT_REQ, idx)
                root_mapping[name] = input_reqfile
                dist_info = _build_root_metadata(roots, name)
                fake_reqs.append(pkg_resources.Requirement(name))
                results.add_dist(dist_info,
                                 input_reqfile)
                idx += 1
            results.add_dist(_build_root_metadata(fake_reqs, ROOT_REQ), ROOT_REQ)
        else:
            roots = utils.reqs_from_files(input_reqfiles)
            results.add_dist(_build_root_metadata(roots, ROOT_REQ), ROOT_REQ)

        with closing(qer.pypi.start_session()) as session:
            qer.compile.compile_roots(root_req, ROOT_REQ, dists=results,
                                      toplevel=root_req, index_url=index_url, session=session,
                                      wheeldir=wheeldir)

            lines = sorted(_generate_lines(results, constraint_results, root_mapping), key=string.lower)
            print('\n'.join(lines))
    except qer.pypi.NoCandidateException as ex:
        _generate_no_candidate_display(ex, results, constraint_results, root_mapping)

    if delete_wheeldir:
        shutil.rmtree(wheeldir)

def compile_main():
    logging.basicConfig(level=logging.ERROR)

    parser = argparse.ArgumentParser()
    parser.add_argument('requirement_files', nargs='+', help='Input requirements files')
    parser.add_argument('-i', '--index-url', type=str, default=None)
    parser.add_argument('-c', '--constraints', action='append',
                        help='Contraints files. Not included in final compilation')
    parser.add_argument('-w', '--wheel-dir', type=str, default=None)
    parser.add_argument('-n', '--no-combine', default=False, action='store_true',
                        help='Keep input requirement file sources separate to '
                             'improve errors and output (slower)')

    args = parser.parse_args()
    run_compile(args.requirement_files, args.constraints if args.constraints else None,
                args.index_url, args.wheel_dir, args.no_combine)


if __name__ == '__main__':
    compile_main()
